@@ -1,13 +1,23 @@
+from __future__ import annotations
+from typing import TYPE_CHECKING
+
+if TYPE_CHECKING:
+    import zarr
+
 from typing import Union
 from os import PathLike
 import os
 from warnings import warn
+from collections.abc import MutableMapping
 
 import numpy as np
 import h5py
 import anndata as ad
 from anndata import AnnData
+
+# from anndata.compat import _read_hdf5_attribute  # 0.8
 from pathlib import Path
+from scipy import sparse
 
 from mudata import MuData
 from .file_backing import MuDataFileManager, AnnDataFileManager
@@ -42,6 +52,9 @@ def _write_h5mu(file: h5py.File, mdata: MuData, write_data=True, **kwargs):
     write_attribute(file, "obsmap", mdata.obsmap, dataset_kwargs=kwargs)
     write_attribute(file, "varmap", mdata.varmap, dataset_kwargs=kwargs)
 
+    attrs = file.attrs
+    attrs["axis"] = mdata.axis
+
     mod = file.require_group("mod")
     for k, v in mdata.mod.items():
         group = mod.require_group(k)
@@ -67,11 +80,14 @@ def _write_h5mu(file: h5py.File, mdata: MuData, write_data=True, **kwargs):
         write_attribute(group, "uns", adata.uns, dataset_kwargs=kwargs)
 
         attrs = group.attrs
-        attrs["encoding-type"] = "AnnData"
+        attrs["encoding-type"] = "anndata"
         attrs["encoding-version"] = __anndataversion__
         attrs["encoder"] = "mudata"
         attrs["encoder-version"] = __version__
 
+    mod_attrs = mod.attrs
+    mod_attrs["mod-order"] = list(mdata.mod.keys())
+
     attrs = file.attrs
     attrs["encoding-type"] = "MuData"
     attrs["encoding-version"] = __mudataversion__
@@ -83,6 +99,104 @@ def _write_h5mu(file: h5py.File, mdata: MuData, write_data=True, **kwargs):
         mdata.update()
 
 
+def write_zarr(
+    store: Union[MutableMapping, str, Path],
+    data: Union[MuData, AnnData],
+    chunks=None,
+    write_data=True,
+    **kwargs,
+):
+    """
+    Write MuData or AnnData object to the Zarr store
+
+    Matrices - sparse or dense - are currently stored as they are.
+    """
+    import zarr
+    from anndata._io.utils import write_attribute
+    from anndata._io.zarr import write_zarr as anndata_write_zarr
+    from .. import __version__, __mudataversion__, __anndataversion__
+
+    if isinstance(data, AnnData):
+        adata = data
+        anndata_write_zarr(store, adata, chunks=chunks, **kwargs)
+    elif isinstance(data, MuData):
+        if isinstance(store, Path):
+            store = str(store)
+        file = zarr.open(store, mode="w")
+        mdata = data
+        write_attribute(
+            file,
+            "obs",
+            mdata.strings_to_categoricals(mdata._shrink_attr("obs", inplace=False)),
+            dataset_kwargs=kwargs,
+        )
+        write_attribute(
+            file,
+            "var",
+            mdata.strings_to_categoricals(mdata._shrink_attr("var", inplace=False)),
+            dataset_kwargs=kwargs,
+        )
+        write_attribute(file, "obsm", mdata.obsm, dataset_kwargs=kwargs)
+        write_attribute(file, "varm", mdata.varm, dataset_kwargs=kwargs)
+        write_attribute(file, "obsp", mdata.obsp, dataset_kwargs=kwargs)
+        write_attribute(file, "varp", mdata.varp, dataset_kwargs=kwargs)
+        write_attribute(file, "uns", mdata.uns, dataset_kwargs=kwargs)
+
+        write_attribute(file, "obsmap", mdata.obsmap, dataset_kwargs=kwargs)
+        write_attribute(file, "varmap", mdata.varmap, dataset_kwargs=kwargs)
+
+        attrs = file.attrs
+        attrs["axis"] = mdata.axis
+
+        mod = file.require_group("mod")
+        for k, v in mdata.mod.items():
+            group = mod.require_group(k)
+
+            adata = mdata.mod[k]
+
+            adata.strings_to_categoricals()
+            if adata.raw is not None:
+                adata.strings_to_categoricals(adata.raw.var)
+
+            if write_data:
+                if chunks is not None and not isinstance(adata.X, sparse.spmatrix):
+                    write_attribute(
+                        group, "X", adata.X, dataset_kwargs=dict(chunks=chunks, **kwargs)
+                    )
+                else:
+                    write_attribute(group, "X", adata.X, dataset_kwargs=kwargs)
+            if adata.raw is not None:
+                write_zarr_raw(group, "raw", adata.raw)
+
+            write_attribute(group, "obs", adata.obs, dataset_kwargs=kwargs)
+            write_attribute(group, "var", adata.var, dataset_kwargs=kwargs)
+            write_attribute(group, "obsm", adata.obsm, dataset_kwargs=kwargs)
+            write_attribute(group, "varm", adata.varm, dataset_kwargs=kwargs)
+            write_attribute(group, "obsp", adata.obsp, dataset_kwargs=kwargs)
+            write_attribute(group, "varp", adata.varp, dataset_kwargs=kwargs)
+            write_attribute(group, "layers", adata.layers, dataset_kwargs=kwargs)
+            write_attribute(group, "uns", adata.uns, dataset_kwargs=kwargs)
+
+            attrs = group.attrs
+            attrs["encoding-type"] = "anndata"
+            attrs["encoding-version"] = __anndataversion__
+            attrs["encoder"] = "mudata"
+            attrs["encoder-version"] = __version__
+
+        mod_attrs = mod.attrs
+        mod_attrs["mod-order"] = list(mdata.mod.keys())
+
+        attrs = file.attrs
+        attrs["encoding-type"] = "MuData"
+        attrs["encoding-version"] = __mudataversion__
+        attrs["encoder"] = "mudata"
+        attrs["encoder-version"] = __version__
+
+        # Restore top-level annotation
+        if not mdata.is_view or not mdata.isbacked:
+            mdata.update()
+
+
 def write_h5mu(filename: PathLike, mdata: MuData, **kwargs):
     """
     Write MuData object to the HDF5 file
@@ -159,7 +273,7 @@ def write_h5ad(filename: PathLike, mod: str, data: Union[MuData, AnnData]):
         write_attribute(fmd, "uns", adata.uns)
 
         attrs = fmd.attrs
-        attrs["encoding-type"] = "AnnData"
+        attrs["encoding-type"] = "anndata"
         attrs["encoding-version"] = __anndataversion__
         attrs["encoder"] = "muon"
         attrs["encoder-version"] = __version__
@@ -184,6 +298,23 @@ def write_h5ad_raw(f, key, raw, **kwargs):
     write_attribute(f, f"{key}/varm", raw.varm, dataset_kwargs=kwargs)
 
 
+def write_zarr_raw(f, key, raw, **kwargs):
+    """
+    Replicates write_raw() in anndata/_io/zarr.py but allow
+    to write raw slots to modalities inside .zarr stores
+    """
+    from anndata._io.zarr import write_attribute
+    from anndata._io.utils import EncodingVersions
+
+    group = f.create_group(key)
+    group.attrs["encoding-type"] = "raw"
+    group.attrs["encoding-version"] = EncodingVersions.raw.value
+    group.attrs["shape"] = raw.shape
+    write_attribute(f, f"{key}/X", raw.X, dataset_kwargs=kwargs)
+    write_attribute(f, f"{key}/var", raw.var, dataset_kwargs=kwargs)
+    write_attribute(f, f"{key}/varm", raw.varm, dataset_kwargs=kwargs)
+
+
 def write(filename: PathLike, data: Union[MuData, AnnData]):
     """
     Write MuData or AnnData to an HDF5 file
@@ -283,15 +414,111 @@ def read_h5mu(filename: PathLike, backed: Union[str, bool, None] = None):
                 for m in gmods.keys():
                     ad = _read_h5mu_mod(gmods[m], manager, backed not in (None, False))
                     mods[m] = ad
+
+                mod_order = None
+                if "mod-order" in gmods.attrs:
+                    mod_order = gmods.attrs["mod-order"]
+                # TODO: use in v0.8
+                # mod_order = _read_hdf5_attribute(k, "mod-order")
+                if mod_order is not None and all([m in gmods for m in mod_order]):
+                    mods = {k: mods[k] for k in mod_order}
+
                 d[k] = mods
             else:
                 d[k] = read_attribute(f[k])
 
+        if "axis" in f.attrs:
+            d["axis"] = f.attrs["axis"]
+
     mu = MuData._init_from_dict_(**d)
     mu.file = manager
     return mu
 
 
+def read_zarr(store: Union[str, Path, MutableMapping, zarr.Group]):
+    """\
+    Read from a hierarchical Zarr array store.
+    Parameters
+    ----------
+    store
+        The filename, a :class:`~typing.MutableMapping`, or a Zarr storage class.
+    """
+    import zarr
+    from anndata._io.zarr import (
+        read_attribute,
+        read_zarr as anndata_read_zarr,
+        read_dataframe,
+        _read_legacy_raw,
+        _clean_uns,
+    )
+
+    if isinstance(store, Path):
+        store = str(store)
+
+    f = zarr.open(store, mode="r")
+    d = {}
+    if "mod" not in f.keys():
+        return anndata_read_zarr(store)
+
+    manager = MuDataFileManager()
+    for k in f.keys():
+        if k in {"obs", "var"}:
+            d[k] = read_dataframe(f[k])
+        if k == "mod":
+            mods = {}
+            gmods = f[k]
+            for m in gmods.keys():
+                ad = _read_zarr_mod(gmods[m], manager)
+                mods[m] = ad
+            d[k] = mods
+        else:  # Base case
+            d[k] = read_attribute(f[k])
+
+    mu = MuData._init_from_dict_(**d)
+    mu.file = manager
+
+    return mu
+
+
+def _read_zarr_mod(g: zarr.Group, manager: MuDataFileManager = None, backed: bool = False) -> dict:
+    import zarr
+    from anndata._io.zarr import read_attribute, read_dataframe, _read_legacy_raw
+    from anndata import Raw
+
+    d = {}
+
+    for k in g.keys():
+        if k in ("obs", "var"):
+            d[k] = read_dataframe(g[k])
+        elif k == "X":
+            X = g["X"]
+            if isinstance(X, zarr.Group):
+                dtype = X["data"].dtype
+            elif hasattr(X, "dtype"):
+                dtype = X.dtype
+            else:
+                raise ValueError()
+            d["dtype"] = dtype
+            if not backed:
+                d["X"] = read_attribute(X)
+        elif k != "raw":
+            d[k] = read_attribute(g[k])
+    ad = AnnData(**d)
+    if manager is not None:
+        ad.file = AnnDataFileManager(ad, os.path.basename(g.name), manager)
+
+    raw = _read_legacy_raw(
+        g,
+        d.get("raw"),
+        read_dataframe,
+        read_attribute,
+        attrs=("var", "varm") if backed else ("var", "varm", "X"),
+    )
+    if raw:
+        ad._raw = Raw(ad, **raw)
+    return ad
+
+
 def _read_h5mu_mod(
     g: "h5py.Group", manager: MuDataFileManager = None, backed: bool = False
 ) -> dict:
